/**************************************************************************\
  class BPatches
  ==============
\**************************************************************************/

#include "BPatches.h"

using namespace std;

// Constructors and destructors
//==========================================================================
BPatches::BPatches(): PointGroup(1)
{ }

BPatches::~BPatches() {}

const string& BPatches::Name()
{
    static const string name("bicubic patch");
    return name;
}

//==========================================================================
// Reading
//==========================================================================

// Read bicubic patches from povray file
//==========================================================================
bool BPatches::ReadPovray(InputStream& is)
{
    static const char* punct[5]={"{","}","<",">",","};
    static const vector<string> punctuation(punct,punct+5);
    static const char* ident[5]=
    {"bicubic_patch","type","flatness","u_steps","v_steps"};
    static const vector<string> identifiers(ident,ident+5);

    Parser reader(is, punctuation, identifiers);
    reader.SkipStrings();

    Parser::Token token;
    Point pnt;

    while(true)
    {
        if(!reader.ReadUntil(Parser::Token(Parser::IDENTIFIER,
                                           "bicubic_patch")))
        {
            if(patches.size()>0) Report();
            else Options::PrintErrorMsg("No bicubic patches found\n");
            return patches.size()>0;
        }

        if(!reader.Expect(Parser::Token(Parser::PUNCTUATION, "{"),
                          "after bicubic_patch"))
          return false;

        unsigned ind=patches.size(), pind=0;
        patches.resize(ind+1);

        unsigned bracketCnt=1;
        while(bracketCnt>0)
        {
            token = reader.NextToken();
            if(token.Type() == Parser::END_OF_FILE)
            { Options::PrintErrorMsg("Missing '}'\n"); return false; }

            if(token.Type() == Parser::PUNCTUATION)
            {
                switch(token.Data()[0])
                { case '{': bracketCnt++; break;
                  case '}': bracketCnt--; break;
                  case '<':
                      if(pind<16)
                      {
                        if(!reader.ReadPoint(pnt, true))
                        { Options::PrintErrorMsg(
                              "Invalid vector at line %u\n", is.LineNumber());
                          return false;
                        }
                        patches[ind].p[pind] = AddPoint(pnt);
                        pind++;
                        if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,
                                                        ">"), ""))
                          return false;
                      }
                      break;
                }
            }
            else if(token.Type() == Parser::IDENTIFIER)
            {
                unsigned i;
                for(i=1; i<5; i++)
                  if(token.Data() == identifiers[i]) break;
                if(i<5)
                {
		    token = reader.NextToken();
                    if(token.Type() != Parser::FLOAT_NUMBER)
                    { Options::PrintErrorMsg(
                        "Error at line %u: Float number expected after %s\n",
                        is.LineNumber(), ident[i]);
                      return false;
                    }
                    switch(i)
                    { case 1: patches[ind].type=token.DataAsInt(); break;
                      case 2: patches[ind].flatness=token.DataAsDouble(); break;
                      case 3: patches[ind].u_steps=token.DataAsInt(); break;
                      case 4: patches[ind].v_steps=token.DataAsInt(); break;
                    }
                }
            }
        }
    }
}

void BPatches::Report()
{
    Options::PrintMsg("Found %u bicubic patches and %u points.\n",
                      patches.size(), PointCount());
}

// Read bicubic patches from PCM file
//==========================================================================
bool BPatches::ReadPCM(InputStream& is)
{
    static const char* punct[1]={","};
    static const vector<string> punctuation(punct,punct+1);
    static const vector<string> identifiers;

    Parser reader(is, punctuation, identifiers);

    ReadPoints(reader);

    unsigned cnt = reader.NextToken().DataAsInt();
    patches.resize(cnt);
    if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,","),""))
      return false;

    unsigned otype=0, ou_steps=3, ov_steps=3;
    double oflatness=0;

    for(unsigned i=0; i<cnt; i++)
    {
        int type = reader.NextToken().DataAsInt();
        if(type == -1)
        { patches[i].type = otype;
          patches[i].flatness = oflatness;
          patches[i].u_steps = ou_steps;
          patches[i].v_steps = ov_steps;
        }
        else
        { otype = patches[i].type = type;
          if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,","),""))
            return false;
	  oflatness = patches[i].flatness = reader.NextToken().DataAsDouble();
          if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,","),""))
            return false;
	  ou_steps = patches[i].u_steps = reader.NextToken().DataAsInt();
          if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,","),""))
            return false;
	  ov_steps = patches[i].v_steps = reader.NextToken().DataAsInt();
        }
        if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,","),""))
          return false;

        for(unsigned j=0; j<16; j++)
        {
            patches[i].p[j] = GivePointPtr(reader.NextToken().DataAsInt());
            if(!reader.Expect(Parser::Token(Parser::PUNCTUATION,","),""))
              return false;
        }
    }

    Report();
    return true;
}

//==========================================================================
// Writing
//==========================================================================

// Write bicubic patches to povray file
//==========================================================================
bool BPatches::WritePovray(ostream& os)
{
    for(unsigned i=0; i<patches.size(); i++)
    {
        os << "bicubic_patch{type " << patches[i].type
           << " flatness " << patches[i].flatness
           << " u_steps " << patches[i].u_steps
           << " v_steps " << patches[i].v_steps;
        for(unsigned j=0; j<16; j++)
        {
            if(j%4==0) os << endl << " ";
            os << "<"; patches[i].p[j]->Print(os); os << ">";
        }
        os << "}" << endl;
    }

    return true;
}

bool BPatches::WritePovray2(ostream& os) { return WritePovray(os); }

// Write bicubic patches to PCM file
//==========================================================================
bool BPatches::WritePCM(ostream& os)
{
    unsigned otype=unsigned(~0), ou_steps=0, ov_steps=0;
    double oflatness=0;

    os << "\"PCM2\"," << endl;
    WritePoints(os);
    os << patches.size() << "," << endl;
    for(unsigned i=0; i<patches.size(); i++)
    {
	if(otype     == patches[i].type &&
           oflatness == patches[i].flatness &&
           ou_steps  == patches[i].u_steps &&
           ov_steps  == patches[i].v_steps)
          os << "-1,";
        else
          os << patches[i].type << "," << patches[i].flatness << ","
             << patches[i].u_steps << "," << patches[i].v_steps << ","
             << endl;
        for(unsigned j=0; j<16; j++)
        {
            os << patches[i].p[j]->Index() << ",";
        }
        os << endl;

        otype = patches[i].type;
        oflatness = patches[i].flatness;
        ou_steps = patches[i].u_steps;
        ov_steps = patches[i].v_steps;
    }
    return true;
}

