/**************************************************************************\
  class Options
  =============
\**************************************************************************/

#include "Options.h"

#include <cstdlib>
#include <cstdarg>
#include <cstdio>
using namespace std;

#ifdef WINMESHCOMP
#pragma hdrstop
#include "winmc/output.h"
#else
static void Print(char* format, va_list arguments)
{ vfprintf(stderr, format, arguments);
}
#endif

void Options::PrintMsg(char* format,...)
{
    if(Quiet()) return;
    va_list argptr;
    va_start(argptr, format);
    Print(format, argptr);
    va_end(argptr);
}
void Options::PrintErrorMsg(char* format,...)
{
    va_list argptr;
    va_start(argptr, format);
    Print(format, argptr);
    va_end(argptr);
}

static double RoundVal(unsigned roundDec)
{
    unsigned i;
    double f=1;

    if(roundDec>0)
        for(i=0;i<roundDec;i++) f*=10;
    return f;
}

//==========================================================================
// Options parsing
//==========================================================================
unsigned Options::ParseOptions(const vector<string>& commandLine,
                               istream** inFile)
{
    /* Set default options */
    overWrite=round=printPov=normalize=optimize=false;
    inputIsPCM=inputIsBP=quiet=useEpsilon=printMesh2=subMeshBoundaries=false;
    extraComp=false;
    dontRemove=true;
    roundDec=5;
    normalLength=10;
    errorTolerance=.001;
    roundVal=1;

    /* Check options */
    unsigned i=1;
    while(commandLine.size()>i &&
          commandLine[i].size()>1 && commandLine[i][0]=='-')
    {
        switch(commandLine[i][1])
        {
          case 'y': overWrite=true;
                    break;
          case 'x': extraComp=true;
                    break;
          case 'r': round=true;
                    if(commandLine[i].size() > 2)
                        roundDec=atoi(commandLine[i].substr(2).c_str());
                    roundVal=RoundVal(roundDec);
                    break;
          case 'p': printPov=true;
                    if(commandLine[i].size() > 2 && commandLine[i][2] == '2')
                        printMesh2=true;
                    break;
          case 'n': normalize=true;
                    if(commandLine[i].size() > 2)
                        normalLength=atof(commandLine[i].substr(2).c_str());
                    break;
          case 'o': optimize=true;
                    if(commandLine[i].size() > 2)
                        errorTolerance=atof(commandLine[i].substr(2).c_str());
                    break;
          case 'i': *inFile=&cin; i++;
                    return i;
          case 'c': inputIsPCM=true;
                    break;
          case 'd': dontRemove=false;
                    break;
          case 'b': inputIsBP=true;
                    break;
          case 'q': quiet=true;
                    break;
          case 'e': useEpsilon=true;
                    if(commandLine[i].size() > 2)
                        epsilon=atof(commandLine[i].substr(2).c_str());
                    break;
          case 'u': noMeshUnion=true;
                    break;
          case 'm': ignoreMesh=true;
                    break;
          default: cerr << "Invalid option: " << commandLine[i] << endl;
                   return 0;
        }
        i++;
    }
    return i;
}

//==========================================================================
// Transformations parsing
//==========================================================================
static bool IsFloat(const char* f)
{
    char* end;
    if(*f==0) return false;
    strtod(f,&end);
    if(*end!=0 && *end!=',') return false;
    return true;
}

static bool IsFloatList(const char* l,int n)
{
    int j1=0,j2,c;
    n--;
    for(c=0;c<n;c++)
    {
        j2=j1;
        while(l[j2] && l[j2]!=',') j2++;
        if(l[j2]==0) break;
        if(!IsFloat(&l[j1])) break;
        j1=j2+1;
    }
    if(c<n || !IsFloat(&l[j1])) return false;
    return true;
}

static inline bool StrNCmp(const string& s1, const string& s2, unsigned n)
{
    return s1.substr(0,n)==s2.substr(0,n);
}

bool Options::ParseTransformations(
    const vector<string>& commandLine, unsigned index)
{
    vector<string>::const_iterator iter = commandLine.begin();
    for(unsigned n=0;n<index;n++) iter++;
    transformations = vector<string>(iter,commandLine.end());

    unsigned i;
    for(i=0; i<TransfCnt(); i++)
    {
        if(transformations[i].size()<2 || transformations[i][0]!='-')
            break;
        switch(transformations[i][1])
        { case 's': if(transformations[i].size()==7 &&
                       StrNCmp(transformations[i].substr(1), "swap", 4))
                    {
                        char c1=transformations[i][5],
                             c2=transformations[i][6];
                        if((c1!='x' && c1!='y' && c1!='z') ||
			   (c2!='x' && c2!='y' && c2!='z'))
                        {
                            PrintMsg("Invalid swap parameter: %s\n",
                                     transformations[i].c_str());
                            return false;
                        }
                    }
                    else
                    {
                        if(!IsFloat(transformations[i].substr(2).c_str()))
                        {
                            PrintMsg("Invalid scale parameter: %s\n",
                                     transformations[i].c_str());
                            return false;
                        }
                    }
                    break;
          case 'c': if(transformations.size()>2 &&
                       (transformations[i][2]!='b' ||
                        transformations[i].size()!=3))
                    {
                        PrintMsg("Invalid center parameter: %s\n",
                                 transformations[i].c_str());
                        return false;
                    }
                    break;
          case 'b': if(transformations[i].size()>2 &&
                       !(((transformations[i][2]=='-' ||
                           transformations[i][2]=='+') &&
                          transformations[i].size()==3) ||
                         IsFloat(transformations[i].substr(2).c_str())))
                    {
                        PrintMsg("Invalid scale to bound parameter: %s\n",
                                 transformations[i].c_str());
                        return false;
                    }
                    if(transformations[i][2]=='+')
                      subMeshBoundaries=true;
                    break;
          case 'r': if(!IsFloatList(transformations[i].substr(2).c_str(),7))
                    {
                        PrintMsg("Invalid rotate parameters: %s\n",
                                 transformations[i].c_str());
                        return false;
                    }
                    break;
          case 't': if(!IsFloatList(transformations[i].substr(2).c_str(),3))
                    {
                        PrintMsg("Invalid translate parameters: %s\n",
                                 transformations[i].c_str());
                        return false;
                    }
                    break;
          default: goto ErroneousParam;
        }
    }
ErroneousParam:
    if(i<TransfCnt())
    {
        PrintMsg("Invalid transformation: %s\n",transformations[i].c_str());
        return false;
    }
    return true;
}


//==========================================================================
// Print syntax
//==========================================================================
void Options::PrintSyntax()
{
#ifndef WINMESHCOMP
    cout <<
"POV-Ray triangle mesh compressor v2.1 by Warp.\n"
"\n"
"Usage: meshcomp [<options>] <infile> [<outfile>] [<transformations>]\n"
"\n"
"Options:\n"
"  -b    : Input is a Bicubic Patch instead of a triangle mesh\n"
"  -c    : Assume that input file is a PCM file instead of a POV-Ray file\n"
"  -y    : Overwrite destination file without confirmation\n"
"  -i    : Read from stdin instead of <infile> (assume <infile>=stdin).\n"
"          NOTE: This parameter MUST be the last option\n"
"  -p    : Output ordinary POV-Ray file instead of PCM file\n"
"  -p2   : Output in uv-pov mesh2 format (only if input is a mesh)\n"
"  -r[n] : Round float values to the nth decimal. If n is not specified,\n"
"          the default value of 5 is used\n"
"  -d    : Try to remove degenerate triangles\n"
"  -e[f] : Set epsilon value to f (by default no epsilon value is used).\n"
"          Example: -e0.00001\n"
"  -u    : Transform meshes independently\n"
"  -m    : Ignore keyword 'mesh'\n"
"  -x    : Use extra compression\n"
"  -q    : Quiet mode. No status text is output to stderr.\n"
"\n"
"These options work only with smooth triangles:\n"
"\n"
"  -n[f] : Normalize all normal vectors (make them equal in length) with\n"
"          length 'f' (a float value). If 'f' is not specified, the default\n"
"          value of 10 is used. (If '-r' option is specified, the values\n"
"          are also rounded)\n"
"  -o[f] : Try to optimize normal vectors (see documentation) with error\n"
"          tolerance 'f'. If 'f' is not specified, the default value of\n"
"          0.001 is used\n"
"\n"
"If <outfile> is not specified, the file is outputted to stdout.\n"
"\n"
"Transformations:\n"
"  -s<float>   : Scale. Multiplies all points by <float>. Example: -s2.5\n"
"  -t<x>,<y>,<z>\n"
"              : Translate. Translates the mesh by (<x>,<y>,<z>) (float\n"
"                values) from the current position. Example: -t1,-5,2.5\n"
"  -c          : Center. Translates all points so that the mesh is centered\n"
"                to <0,0,0>.\n"
"  -cb         : Center Bottom. Translates all points so that the bottom of\n"
"                the mesh is centered to <0,0,0> (the negative y coordinate\n"
"                is considered the bottom).\n"
"  -b[<float>] : scale to Bound. Centers the mesh and then scales all points\n"
"                so that the mesh gets bounded by a box of size <float>. This\n"
"                is equivalent to applying -c and then -s with an appropriate\n"
"                value. If <float> is not specified, the default value of 2\n"
"                is used (ie. the box extends from <-1,-1,-1> to <1,1,1>).\n"
"  -b-         : outputs the minimum and maximum values (ie. Boundaries) of\n"
"                the mesh to stderr (later transformations may change this)\n"
"  -b+         : Like -b- but outputs also the boundaries of all the\n"
"                submeshes (identical to -b- if option -u specified)\n"
"  -swap(x|y|z)(x|y|z)\n"
"              : Swap coordinates. For example, -swapxz swaps all x and z\n"
"                coordinates. If you specify the same coordinate as both\n"
"                parameters, that coordinate is negated, ie. -swapxx will\n"
"                change the sign of all x coordinates (the mesh is mirrored\n"
"                about the yz-plane).\n"
"  -r<angle>,<x1>,<y1>,<z1>,<x2>,<y2>,<z2>\n"
"              : Rotate. The parameter is followed by 7 float values separated\n"
"                by commas. The first value is the rotation angle, and the next\n"
"                6 values define two points. The points are rotated <angle>\n"
"                radians around the axis defined by the two points. Example:\n"
"                -r3.14,0,0,0,10,-5,2.5\n"
"\n"
"Example: meshcomp -r4 -o0.1 dolphin.inc dolphin.pcm -swapyz -b10 -cb\n"
"\n"
"Send comments, suggestions and bug reports to warp@iki.fi\n";
#endif
}


// Static variables:
bool Options::overWrite, Options::round, Options::printPov,
     Options::normalize, Options::optimize;
bool Options::inputIsPCM, Options::inputIsBP, Options::quiet,
     Options::useEpsilon, Options::dontRemove, Options::printMesh2;
bool Options::noMeshUnion, Options::ignoreMesh, Options::subMeshBoundaries;
bool Options::extraComp;
unsigned Options::roundDec;
double Options::epsilon, Options::normalLength, Options::errorTolerance,
       Options::roundVal;

vector<string> Options::transformations;

