/**************************************************************************\
  class Parser
  ============
\**************************************************************************/

#include "Parser.h"

#include <algorithm>
#include <cctype>

// Constructor
//========================================================================
Parser::Parser(InputStream& is,
               const vector<string>& punct,
               const vector<string>& identifiers):
    is(is), punctuation(punct),
    identifiers(identifiers.begin(), identifiers.end()),
    skipStrings(false), nextToken()
{
    sort(punctuation.begin(), punctuation.end());
}

// Skip string from given stream (the starting " has been already read)
//========================================================================
static inline void SkipString(InputStream& is)
{
    char c, prev=0;
    while(is.good())
    {
        is >> c;
        if(c == '"' && prev!='\\') return;
        prev = c;
    }
}

// Read identifier from given stream to given string
//========================================================================
static inline void ReadIdentifier(InputStream& is, string& item)
{
    char c;
    while(is.good())
    {
        c = is.peek();
        if(!isalnum(c) && c!='_') return;
        item += is.get();
    }
}

// Read next token from stream
//========================================================================
Parser::Token Parser::NextToken()
{
    if(nextToken.Type() != NONE)
    {
        Token ret = nextToken;
        nextToken.Set(NONE);
        return ret;
    }

    char c;
    string item;

    while(is.good())
    {
      // Check for punctuation (currently only tokens of 1 character are valid)
      is >> c;
      for(unsigned i=0; i<punctuation.size(); i++)
        if(punctuation[i][0] == c)
          return Token(PUNCTUATION, string(1,c));

      // Check for string:
      if(c == '"')
      {
	  if(skipStrings) { SkipString(is); continue; }
          string tmp;
          do
          {
              is.getline(tmp, '"');
	      item += tmp;
          } while(is.good() && tmp.size()>1 && tmp[tmp.size()-2]=='\\');
          item.erase(item.size()-1);
          return Token(STRING, item);
      }

      // Check for identifier
      if(isalpha(c) || c=='_')
      {
	  item += c;
          ReadIdentifier(is, item);
          if(set<string>::const_iterator(identifiers.find(item)) !=
             identifiers.end())
            return Token(IDENTIFIER, item);
          item.erase();
          continue;
      }

      // Check for number
      if(isdigit(c) || c=='-' || c=='.')
      {
          bool digitread = isdigit(c), pointread = (c=='.'), eread = false;
          item += c;
          while(is.good())
          {
              c = is.peek();
              if(!((c=='e'||c=='E') && digitread && !eread) &&
                 ((!isdigit(c) && c!='.') || (c=='.' && pointread)))
              {
		  if(digitread)
                    return Token(FLOAT_NUMBER, item);
                  else
                  { item.erase(); break; }
              }
              item += is.get();
              if(isdigit(c)) digitread=true;
              else if(c=='.') pointread=true;
              else if(c=='e' || c=='E')
              {
                  eread=true;
                  if(is.peek() == '-')
                    item += (char)is.get();
                  else if(is.peek() == '+')
                    is.get();
              }
          }
      }
    }

    return Token(END_OF_FILE);
}

// Read input stream until the specified token appears
//========================================================================
bool Parser::ReadUntil(const Token& token)
{
    Token tmp;
    while(true)
    {
        tmp=NextToken();
        if(tmp.Type() == END_OF_FILE) return false;
        if(tmp == token) return true;
    }
}

// Read a 3D-point from input stream
//========================================================================
bool Parser::ReadPoint(Point& pnt, bool expand)
{
    if(PeekToken().Type() != FLOAT_NUMBER)
    { pnt.Index(0); return false; }
    pnt.x(NextToken().DataAsDouble());
    if(PeekToken().Type()==PUNCTUATION && PeekToken().Data()==",")
      NextToken();

    if(PeekToken().Type() != FLOAT_NUMBER)
    {
        if(expand &&
           PeekToken().Type()==PUNCTUATION && PeekToken().Data()==">")
        {
            pnt.y(0); pnt.z(0); return true;
        }
        pnt.Index(1); return false;
    }
    pnt.y(NextToken().DataAsDouble());
    if(PeekToken().Type()==PUNCTUATION && PeekToken().Data()==",")
      NextToken();

    if(PeekToken().Type() != FLOAT_NUMBER)
    {
        if(expand &&
           PeekToken().Type()==PUNCTUATION && PeekToken().Data()==">")
        {
            pnt.z(0); return true;
        }
        pnt.Index(2); return false;
    }
    pnt.z(NextToken().DataAsDouble());

    return true;
}

// Read a block out of the stream
//========================================================================
bool Parser::ReadBlock(unsigned bracketCnt,
                       char openBracket, char closeBracket)
{
    char c;
    if(nextToken.Type() != NONE)
    {
        if(nextToken.Data()[0] == openBracket) bracketCnt++;
        else if(nextToken.Data()[0] == closeBracket) bracketCnt--;
        nextToken.Set(NONE);
    }
    while(bracketCnt>0)
    {
        if(!is.good()) return false;
        c = is.get();
        if(c==openBracket) bracketCnt++;
        else if(c==closeBracket) bracketCnt--;
    }
    return true;
}


